/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 */
package org.enhydra.zeus;

import org.enhydra.zeus.result.StreamResult;
import org.enhydra.zeus.util.CapitalizationUtils;

import java.io.IOException;
import java.io.OutputStream;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.Iterator;
import java.util.List;

// JDOM imports
import org.jdom.Attribute;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.output.XMLOutputter;

/**
 * <p>
 *  <code>Marshaller</code> takes a Java <code>Object</code> instance and writes
 *    out an XML representation of that object, with each property (variable)
 *    of the instance and its value.
 * </p>
 *
 * @author Brett McLaughlin
 */
public class Marshaller {

    /**
     * <p>
     *  Simple constructor.
     * </p>
     */
    public Marshaller() {
        // Currently, do nothing
    }
   
    /**
     * <p>
     *  This method is the public entry point for marshalling an object into
     *    an XML instance document.
     * </p>
     *
     * @param obj <code>Object</code> to convert to XML.
     * @param result <code>Result</code> to write XML to.
     * @throws <code>IOException</code> when errors in output occur.
     */
    public void marshall(Object obj, Result result) throws IOException {
        // Root Element is the start of recursion
        Element root = getXMLRepresentation(obj);
        Document doc = new Document(root);
        
        // XXX: For now, I don't know how we would write to anthing but
        //   a StreamResult. So this is a temp hack, for sure!
        if (result instanceof StreamResult) {
            StreamResult streamResult = (StreamResult)result;

            // Use 2 space indentation and line feeds
            XMLOutputter outputter = new XMLOutputter("  ", true);            
            outputter.output(doc, streamResult.getWriter());
            streamResult.getWriter().flush();
        } else {
            throw new IllegalArgumentException("Unmarshaller currently only works with StreamResults.");
        }
    }

    /**
     * <p>
     *  This is the granular portion of binding; a Java <code>Object</code> is
     *    converted into an XML element (in JDOM form), using recursion for any
     *    children.
     * </p>
     * 
     * @param obj <code>Object</code> to get the XML element representation for.
     * @return <code>Element</code> - representation of <code>Object</code> in XML.
     * @throws <code>IOException</code> when errors occur in binding.
     */
    private Element getXMLRepresentation(Object obj) throws IOException {
        Class objectClass = obj.getClass();

        // Get the name of the element for this object
        String objectName = CapitalizationUtils.initialLower(objectClass.getName());

        // If this is an "Impl" class, remove that from the name
        int index = -1;
        if ((index = objectName.indexOf("Impl")) != -1) {
            objectName = objectName.substring(0, index);
        }

        Element element = new Element(objectName);

        Method[] methods = objectClass.getMethods();
        for (int i=0; i<methods.length; i++) {
            // Deal with attributes
            Method method = methods[i];
            String methodName = method.getName();
            if ((methodName.startsWith("get")) && 
                (!methodName.equals("getClass")) &&
                (!methodName.equals("getText")) &&
                (!methodName.endsWith("List"))) {

                // Get the value for this method
                try {
                    Object o = method.invoke(obj, new Object[] { });

                    // For the name, remove the "get" and lower the initial letter
                    String propertyName = 
                        CapitalizationUtils.initialLower(method.getName().substring(3));
                    
                    // Determine if it's primitive by seeing if it's a java.lang type
                    if (o.getClass().getName().startsWith("java.lang.")) {
                        // If it's a primitive, add as an attribute
                        element.addAttribute(propertyName, o.toString());
                    } else {
                        // ... otherwise, recurse and add new element as child
                        element.addContent(getXMLRepresentation(o));
                    }

                } catch (IllegalAccessException e) {
                    throw new IOException(e.getMessage());
                } catch (InvocationTargetException e) {
                    throw new IOException(e.getMessage());
                }
            } else if ((methodName.startsWith("get")) &&
                       (methodName.endsWith("List"))) {
                // Deal with List data
                try {
                    List items = (List)method.invoke(obj, new Object[] { });
                    for (Iterator j = items.iterator(); j.hasNext(); ) {
                        element.addContent(getXMLRepresentation(j.next()));
                    }
                } catch (IllegalAccessException e) {
                    throw new IOException(e.getMessage());
                } catch (InvocationTargetException e) {
                    throw new IOException(e.getMessage());
                }                
            } else if (methodName.equals("getText")) {
                // Deal with Data field
                try {
                    String content = (String)method.invoke(obj, new Object[] { });
                    element.setText(content);
                } catch (IllegalAccessException e) {
                    throw new IOException(e.getMessage());
                } catch (InvocationTargetException e) {
                    throw new IOException(e.getMessage());
                }                    
            }
        }

        return element;
    }
}
