/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 */
package org.enhydra.zeus;

import java.io.IOException;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

// Zeus imports
import org.enhydra.zeus.util.CapitalizationUtils;
import org.enhydra.zeus.util.ClassUtils;

// JDOM imports
import org.jdom.Attribute;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.Namespace;
import org.jdom.input.SAXBuilder;

/**
 * <p>
 *  <code>Unmarshaller</code> takes an XML instance document, which should
 *    conform to some set of XML constraints, and creates a Java object from the XML
 *    document. This object is an instance of the implementation class, which
 *    in turn implements the interface, created using an implementation of
 *    <code>{@link Generator}</code>.
 * </p>
 * <p>
 *  It also assumes that the generated interface and implementation classes
 *    are in the classpath, as it will perform a <code>Class.forName</code> on
 *    the classes.
 * </p>
 *
 * @author Brett McLaughlin
 * @version 1.0
 */
public class Unmarshaller {
    
    /** Store interface/implementation class mappings */
    private Map implClasses;

    /**
     * <p>
     *  Simple constructor.
     * </p>
     */
    public Unmarshaller() {
        implClasses = new HashMap();
    }
    
    /**
     * <p>
     *  For the supplied interface name, this will allow a customized implementation
     *    class to be set. This effectively allows the user to specify their
     *    own implementation class to use instead of the default, Zeus-generated
     *    class.
     * </p>
     *
     * @param interfaceName <code>String</code> name of interface being deal with.
     * @param implClassName <code>String</code> name of class to load when implementations
     *        of <code>interfaceName</code> are needed.
     */
    public void setImplClass(String interfaceName, String implClassName) {
        implClasses.put(interfaceName, implClassName);
    }
    
    /**
     * <p>
     *  For the supplied interface name, this will return the current implementation
     *    class associated with the interface. If no class has been specified, this will
     *    return the default, Zeus-generated class name for the interface.
     * </p>
     *
     * @param interfaceName <code>String</code> name of interface being dealt with.
     * @return <code>String</code> - implementation class for supplied 
     *         <code>interfaceName</code>.
     */
    public String getImplClass(String interfaceName) {
        String implClass = (String)implClasses.get(interfaceName);
        if (implClass == null) {
            // Construct class name for Zeus binding class
            implClass = new StringBuffer()
                .append(interfaceName)
                .append("Impl")
                .toString();
        }
        return implClass;
    }
   
    /**
     * <p>
     *  This method is the public entry point for unmarshalling an object from
     *    an XML instance document.
     * </p><p>
     *  XXX: We need to eventually put in Zeus exceptions here, not IOExceptions.
     * </p>
     *
     * @param instanceURL <code>URL</code> for the instance document.
     * @return <code>Object</code> - the created Java object, or 
     *         <code>null</code> if problems occur in a way that does not
     *         generate an <code>Exception</code>.
     * @throws <code>IOException</code> when errors in binding occur.
     */
    public Object unmarshall(Source sourceXML) throws IOException {
        // Obtain the JDOM Document from the supplied source
        Document doc = sourceXML.getDocument(); 
        Element rootElement = doc.getRootElement();
        Unmarshaller unmarshaller = new Unmarshaller();
            
        // Convert to Java
        return unmarshaller.getJavaRepresentation(rootElement);
    }

    /**
     * <p>
     *  This will take an XML document fragment, starting with the provided
     *    <code>Element</code>, and convert it to a Java representation.
     * </p>
     *
     * @param rootElement <code>Element</code> to begin processing on.
     * @return <code>Object</code> - unmarshalled Java object.
     * @throws <code>IOException</code> - when XML cannot be converted to Java.
     */
    private Object getJavaRepresentation(Element element) 
        throws IOException {           

        Object obj = null;
        String className = null;

        try {
            Namespace ns = element.getNamespace();

            // Class name is root element name, with initial caps, plus "Impl"
            className = getImplClass(CapitalizationUtils.justInitialUpper(element.getName()));

            Class objectClass;

            // Get the class
            objectClass = Class.forName(className);

            // Get a new instance of this class
            obj = objectClass.newInstance();

            // For each attribute, get its name and call mutator
            List attributes = element.getAttributes();
            Method[] methods = objectClass.getMethods(); 

            for (Iterator i = attributes.iterator(); i.hasNext(); ) {
                Attribute att = (Attribute)i.next();

                // Only want attributes for this namespace
                if ((!att.getNamespace().equals(ns)) &&
                    (!att.getNamespace().equals(Namespace.NO_NAMESPACE))) {
                    continue;
                }

                // Determine method to call - no Lists to worry about with attributes
                String methodName = new StringBuffer()
                    .append("set")
                    .append(CapitalizationUtils.justInitialUpper(att.getName()))
                    .toString();

                // Find the method to call, and its parameter type
                for (int j=0; j<methods.length; j++) {
                    if (methods[j].getName().equals(methodName)) {
                        // Since all mutators have one param, get the first one
                        Class[] paramTypes = methods[j].getParameterTypes();
                        Class paramType = paramTypes[0];

                        // Convert the type we have to the correct type
                        Object param = 
                            ClassUtils.getParameter(att.getValue(), paramType);

                        // Invoke the method
                        methods[j].invoke(obj, new Object[] { param });
                    }
                }
            }

            // Now do complex objects
            List children = element.getChildren();

            for (Iterator i = children.iterator(); i.hasNext(); ) {
                Element child = (Element)i.next();               

                // Only want elements for this namespace
                if ((!child.getNamespace().equals(ns)) &&
                    (!child.getNamespace().equals(Namespace.NO_NAMESPACE))) {
                    continue;
                }

                // Determine method to call - start with default name
                String singularMethodName = new StringBuffer()
                    .append("set")
                    .append(CapitalizationUtils.justInitialUpper(child.getName()))
                    .toString();
                String listMethodName = new StringBuffer()
                    .append("add")
                    .append(CapitalizationUtils.justInitialUpper(child.getName()))
                    .toString();              

                // Find the method to call, and its parameter type
                for (int j=0; j<methods.length; j++) {
                    if (methods[j].getName().equals(singularMethodName)) {                                             
                        // Since all mutators have one param, get the first one
                        Class[] paramTypes = methods[j].getParameterTypes();
                        Class paramType = paramTypes[0];

                        // Convert the type we have to the correct type
                        Object param = 
                            getJavaRepresentation(child);

                        // Invoke the method
                        methods[j].invoke(obj, new Object[] { param });
                        break;
                    } else if (methods[j].getName().equals(listMethodName)) {
                        Class[] paramTypes = methods[j].getParameterTypes();
                        Class paramType = paramTypes[0];
                        
                        // Convert the type we have to the correct type
                        Object param =
                            getJavaRepresentation(child);
                        
                        // Invoke the method                        
                        methods[j].invoke(obj, new Object[] { param });
                        break;
                    }
                }                
            }
            
            // Finally, deal with element content            
            String content = element.getTextTrim();
            if ((content != null) && (content.length() > 0)) {
                String methodName = "setText";
                Method method = objectClass.getMethod(methodName, new Class[] { String.class });
                method.invoke(obj, new Object[] { content });               
            }
            
        } catch (ClassNotFoundException e) {
            throw new IOException("The class that the XML instance document " +
                "should be unmarshalled as an instance of, " + className +
                ", must be generated and on the classpath.");
        } catch (InstantiationException e) {
            throw new IOException("Could not create a new instance of the " +
                "class " + className);
        } catch (IllegalAccessException e) {
            throw new IOException("Could not access the class " + className);
        } catch (IllegalArgumentException e) {
            throw new IOException("Could not pass arguments to mutator " +
                "method on new instance.");
        } catch (InvocationTargetException e) {
            throw new IOException("Error in accessing mutator on new " +
                "instance of class.");
        } catch (NoSuchMethodException e) {
            throw new IOException("Error in setting data method on new instance of class.");
        }
        return obj;
    } 

}
